// joystick.c

#include "ntddk.h"
#include "gameport.h"
#include "joystick.h"

typedef struct _DEVICE_EXTENSION {
   PDEVICE_OBJECT    TargetDevice;
   UCHAR             GamePortMask;
} DEVICE_EXTENSION, *PDEVICE_EXTENSION;

#define JOY_DEVICE1_MASK    0x33   // 00110011
#define JOY_DEVICE2_MASK    0xCC   // 11001100
#define JOY_BUTTON1_MASK    0x50   // 01010000
#define JOY_BUTTON2_MASK    0xA0   // 10100000

NTSTATUS JoyStickCreateClose(IN PDEVICE_OBJECT, IN PIRP);
NTSTATUS JoyStickRead(IN PDEVICE_OBJECT, IN PIRP);
VOID JoyStickUnload(IN PDRIVER_OBJECT);
NTSTATUS JoyStickCompletion(IN PDEVICE_OBJECT, IN PIRP Irp, PVOID);

//-----------------------------------------------------------------
NTSTATUS DriverEntry(IN PDRIVER_OBJECT  DriverObject,
                     IN PUNICODE_STRING RegistryPath)
{
    NTSTATUS           status = STATUS_SUCCESS;
    PDEVICE_EXTENSION  deviceExt;
    PDEVICE_OBJECT     deviceObject0, deviceObject1, targetDevice;
    PFILE_OBJECT       fileObject;
    UNICODE_STRING     deviceName, symName0, symName1;

    UNREFERENCED_PARAMETER(RegistryPath);

    KdPrint(("JoyStick!DriverEntry\n"));

    //
    // First, get pointer to gameport device object
    //

    RtlInitUnicodeString(&deviceName, GAMEPORT_DEV_NAME);
    status = IoGetDeviceObjectPointer(&deviceName, 
                    FILE_ANY_ACCESS, &fileObject, &targetDevice);
    if (!NT_SUCCESS(status)) {
        KdPrint(("JoyStick: Get DevObj failed\n"));
        return status;
    }

    ObReferenceObjectByPointer(targetDevice, FILE_ALL_ACCESS,
                               NULL, KernelMode);
    ObDereferenceObject(fileObject);

    //
    // Create exclusive device object to represent joystick0
    //
    
    RtlInitUnicodeString(&deviceName, JOYSTICK0_DEV_NAME);
    RtlInitUnicodeString(&symName0, JOYSTICK0_LINK_NAME);                 
    status = IoCreateDevice(DriverObject,sizeof(DEVICE_EXTENSION),
                            &deviceName, FILE_DEVICE_JOYSTICK, 0,
                            TRUE, &deviceObject0);
    if (!NT_SUCCESS(status)) {
        ObDereferenceObject(targetDevice);
        return status;
    }
    IoCreateSymbolicLink(&symName0, &deviceName);
    deviceObject0->Flags |= DO_BUFFERED_IO;
    deviceObject0->StackSize = targetDevice->StackSize + 1;
    deviceObject0->AlignmentRequirement = 
                        targetDevice->AlignmentRequirement;
    deviceExt = (PDEVICE_EXTENSION)deviceObject0->DeviceExtension;
    deviceExt->TargetDevice = targetDevice;  
    deviceExt->GamePortMask = JOY_DEVICE1_MASK;

    //
    // Create exclusive device object to represent joystick1
    //
    
    RtlInitUnicodeString(&deviceName, JOYSTICK1_DEV_NAME);
    RtlInitUnicodeString(&symName1, JOYSTICK1_LINK_NAME);                 
    status = IoCreateDevice(DriverObject,sizeof(DEVICE_EXTENSION),
                            &deviceName, FILE_DEVICE_JOYSTICK, 0,
                            TRUE, &deviceObject1);
    if (!NT_SUCCESS(status)) {
        IoDeleteSymbolicLink(&symName0);
        IoDeleteDevice(deviceObject0);
        ObDereferenceObject(targetDevice);
        return status;
    }
    IoCreateSymbolicLink(&symName1, &deviceName);
    deviceObject1->Flags |= DO_BUFFERED_IO;
    deviceObject1->StackSize = targetDevice->StackSize + 1;
    deviceObject1->AlignmentRequirement = 
                        targetDevice->AlignmentRequirement;
    deviceExt = (PDEVICE_EXTENSION)deviceObject1->DeviceExtension;
    deviceExt->TargetDevice = targetDevice;  
    deviceExt->GamePortMask = JOY_DEVICE2_MASK;

    //
    // Fill in the dispatch table.
    //

    DriverObject->MajorFunction[IRP_MJ_CREATE] = JoyStickCreateClose;
    DriverObject->MajorFunction[IRP_MJ_CLOSE]  = JoyStickCreateClose;
    DriverObject->MajorFunction[IRP_MJ_READ]   = JoyStickRead;
    DriverObject->DriverUnload                 = JoyStickUnload;
    return status;

} // DriverEntry

//-----------------------------------------------------------------
NTSTATUS JoyStickCreateClose(IN PDEVICE_OBJECT DevObject, 
                             IN PIRP Irp)
{
    KdPrint(("JoyStickCreateClose\n"));

    Irp->IoStatus.Status = STATUS_SUCCESS;
    Irp->IoStatus.Information = 0;
    IoCompleteRequest(Irp, IO_NO_INCREMENT);
    return STATUS_SUCCESS;

} // JoyStickCreateClose

//-----------------------------------------------------------------
NTSTATUS JoyStickRead(IN PDEVICE_OBJECT DevObject, IN PIRP Irp)
{
    PIO_STACK_LOCATION currentIrpStack, nextIrpStack;
    PDEVICE_EXTENSION deviceExt = DevObject->DeviceExtension;

    KdPrint(("JoyStickRead\n"));

    currentIrpStack = IoGetCurrentIrpStackLocation(Irp);
    nextIrpStack = IoGetNextIrpStackLocation(Irp);

    //
    // Validate parameters (gameport expects JOY_DATA struct).
    //
    
    if (currentIrpStack->Parameters.Read.Length
                < sizeof(JOY_DATA)) {
        Irp->IoStatus.Status = STATUS_BUFFER_TOO_SMALL;
        Irp->IoStatus.Information = 0;
        return STATUS_BUFFER_TOO_SMALL;
    }

    //
    // Get the button data by forwarding a 1-byte read to the
    // gameport device with a completion routine registered.
    //

    RtlMoveMemory(nextIrpStack, currentIrpStack,
                  sizeof(IO_STACK_LOCATION));
    nextIrpStack->Parameters.Read.Length = 1;
    nextIrpStack->Parameters.Read.Key = 0;     // private use 

    IoSetCompletionRoutine(Irp, JoyStickCompletion, NULL, TRUE,
                           TRUE, TRUE); 

    return IoCallDriver(deviceExt->TargetDevice, Irp);
                              
} // JoyStickRead

//-----------------------------------------------------------------
NTSTATUS JoyStickCompletion(IN PDEVICE_OBJECT DevObject, 
                            IN PIRP Irp, IN PVOID Context)
{
    PDEVICE_EXTENSION deviceExt = DevObject->DeviceExtension;
    JOY_DATA joyData;
    UCHAR data;

    KdPrint(("JoyStickCompletion\n"));

    if (NT_SUCCESS(Irp->IoStatus.Status)) {
        data = *(PUCHAR)(Irp->AssociatedIrp.SystemBuffer);
        data ^= 0xf0;                    // invert button bits
        data &= 0xf0;                    // clear axis bits        
        data &= deviceExt->GamePortMask; // JoyStickX data only
        joyData.Button1 = (data & JOY_BUTTON1_MASK) ? 1 : 0;
        joyData.Button2 = (data & JOY_BUTTON2_MASK) ? 1 : 0;
        RtlMoveMemory(Irp->AssociatedIrp.SystemBuffer, 
                      &joyData, sizeof(JOY_DATA));
        Irp->IoStatus.Information = sizeof(JOY_DATA);
    }    

    if (Irp->PendingReturned) {
        IoMarkIrpPending(Irp);
    }

    return STATUS_SUCCESS;

} // JoyStickCompletion

//-----------------------------------------------------------------
VOID JoyStickUnload(IN PDRIVER_OBJECT DriverObject)
{
    PDEVICE_OBJECT next = DriverObject->DeviceObject->NextDevice;
    PDEVICE_EXTENSION deviceExt = next->DeviceExtension;
    UNICODE_STRING symName;

    KdPrint(("JoyStickUnload\n"));

    RtlInitUnicodeString(&symName, JOYSTICK0_LINK_NAME);
    IoDeleteSymbolicLink(&symName);
    RtlInitUnicodeString(&symName, JOYSTICK1_LINK_NAME);
    IoDeleteSymbolicLink(&symName);

    ObDereferenceObject(deviceExt->TargetDevice);
    IoDeleteDevice(DriverObject->DeviceObject);
    IoDeleteDevice(next);

} // JoyStickUnload
