/*******************************************************************
***************************  rttest.c  *****************************
********************************************************************
*
*  Test application for Better Windows NT real time frame generator
*
*  Developed by:  Lionell K. Griffith - March 31, 1998
*
*  email:  lgriffith@hughes.net
*
*     Procedure      Description  
*  1. AppFrameProc   Frame timer callback for test application
*  2. main           Main procedure for test application
*
*  NOTE:  winmm.lib must be in the linker library list
*
*******************************************************************/

#define STRICT

#include <stdio.h>
#include <windows.h>
#include "realtime.h"

struct TestStruct
  {
   long Periods;                  /* Number of test Periods       */
   long Frame;                    /* Frame Counter                */
   long Running;                  /* Idle loop run flag           */
   long Error20us;                /* Count of <= 20 us error      */
   long Error50us;                /* Count of <= 50 us error      */
   long Error200us;               /* Count of <= 200 us error     */
   long Error500us;               /* Count of <= 500 us error     */
   double OldTime;                /* Old frame time               */
   double Target;                 /* Target frame period          */
   double Period;                 /* Measured frame period        */
   double Sum;                    /* Sum of periods               */
   double Mean;                   /* Mean period                  */
   double Max;                    /* Maximum period               */
   double Min;                    /* Minimum period               */
  };
typedef struct TestStruct TestType;

/*******************************************************************
**************************  1. AppFrameProc  ***********************
********************************************************************
*
*   Frame timer callback for test application
*
*   AppFrameProc(pData);
*
*   Void pointer to application defined data
*
*******************************************************************/

void AppFrameProc(void *pData)
  {
   TestType *pTest = (TestType*)pData;
   double Time;
   double Error;

   /* Accumulate frame statistics skipping first frame */
   Time = FrameTime();
   pTest->Period = Time-pTest->OldTime;
   Error = fabs(pTest->Period - pTest->Target);
   if ((pTest->Frame > 0) && (pTest->Frame < pTest->Periods + 1))
     {
      pTest->Sum = pTest->Sum + pTest->Period;
      if (pTest->Period > pTest->Max) pTest->Max = pTest->Period;
      if (pTest->Period < pTest->Min) pTest->Min = pTest->Period;
      if (Error < 0.000020) 
        pTest->Error20us = pTest->Error20us + 1;
      else if (Error <= 0.000050) 
        pTest->Error50us = pTest->Error50us + 1;
      else if (Error <= 0.000200) 
        pTest->Error200us = pTest->Error200us + 1;
      else if (Error <= 0.000500) 
        pTest->Error500us = pTest->Error500us + 1;
     }
   pTest->OldTime = Time;
   pTest->Frame = pTest->Frame + 1;
   if (pTest->Frame > pTest->Periods) pTest->Running = FALSE;
  }

/*******************************************************************
**************************  2. main  *******************************
********************************************************************
*
*  Main procedure for test application
*
*******************************************************************/

int main(int argc,char *argv)
  {
   TestType Test;

   /* Initialize Fast Timer */
   FastTimerInit();

   /* Initialize test */
   Test.Frame = 0;
   Test.Sum = 0.0;
   Test.Max = 0.0;
   Test.Min = 10.0;
   Test.Periods = 2500;
   Test.Target = 1.0/60.0;
   Test.Error20us = 0;
   Test.Error50us = 0;
   Test.Error200us = 0;
   Test.Error500us = 0;

   /* Start frame timer */
   SetPriorityClass(GetCurrentProcess(),REALTIME_PRIORITY_CLASS);
   FrameSet(Test.Target,AppFrameProc,&Test);
   Test.OldTime = FrameTime();

   /* Do Idle loop */
   Test.Running = TRUE;
   do
     {
      Sleep(100);
     }
   while (Test.Running);

   /* Terminate Frame Timer */
   FrameKill();
   SetPriorityClass(GetCurrentProcess(),NORMAL_PRIORITY_CLASS);

   /* Output results */
   printf("\nPeriod Statistics\n\n");
   Test.Mean = Test.Sum/(double)Test.Periods;
   printf("Target Frequency                 = %9.3f\n",
          1.0/Test.Target);
   printf("Mean Frequency                   = %9.3f\n",
          1.0/Test.Mean);
   printf("Max Frequency                    = %9.3f\n",
          1.0/Test.Min);
   printf("Min Frequency                    = %9.3f\n",
          1.0/Test.Max);
   printf("Target Period                    = %9.6f\n",
          Test.Target);
   printf("Mean Period                      = %9.6f\n",
          Test.Mean);
   printf("Max Period                       = %9.6f\n",
          Test.Max);
   printf("Min Period                       = %9.6f\n",
          Test.Min);
   printf("Periods                          =  %d\n",
          Test.Periods);
   printf("Periods with error up to 20 us   =  %d\n",
          Test.Error20us);
   printf("Periods with error 20 to 50 us   =  %d\n",
          Test.Error50us);
   printf("Periods with error 50 to 200 us  =  %d\n",
          Test.Error200us);
   printf("Periods with error 200 to 500 us =  %d\n",
          Test.Error500us);
   printf("\n");
   return 0;
  }

/*******************************************************************
************************  End of rttest.c  *************************
*******************************************************************/

